/* $Id: Builder.java,v 2.2 2002/04/26 14:18:16 visnovsky Exp $ */

package SOFA.SOFAnode.Util;

import java.io.*;
import java.util.*;

/**
 * The class for building the automata from a given behavior protocol. A protocol is passed
 * not into constructor, thus, one instance can be used for building any number of automata.
 *
 * @author Stanislav Visnovsky
 * @version 1.0.0
 *
 * @see Machine
 * @see checker
*/
public class Builder {
/**
 * next char for LL(1) parsing
*/
    private char next_char;

/**
 * a reader which contains a behavior protocol to be parsed and built
*/
    private Reader source;

/**
 * next token for LL(1) parsing
*/
    private String next_token;

/**
 * flag for indicating the action token read
*/
    private boolean isTokenID;

/**
 * Test for character to be a letter (case insensitive).
 *
 * @author 	Stanislav Visnovsky
 * @version 	1.0.0
 *
 * @param ch	character to be tested
 * @return	result of the test
 * @see		Builder#nextToken
*/
    private boolean isLetter( char ch ) {
        return ( ch>='a' && ch<='z' ) || ( ch>='A' && ch<='Z' );
    }

/**
 * Read the next token from the reader. In case of action token, indicate it by the isTokenID flag.
 *
 * @author 	Stanislav Visnovsky
 * @version 	1.5.1
 *
 * @return	read token, or token EOF
 * @throws SyntaxErrorException		in case of badly formed protocols
 * @see		Builder#isTokenID
*/
    private String nextToken() throws SyntaxErrorException, IOException {
	String result;
        isTokenID = false;
        if( !source.ready() ) {
            Debug.println( 2, "EOF" );
            return "EOF";
        }
        while( next_char == ' ' )            // skip whitespaces
            if( source.ready() ) next_char = (char)source.read();
            else {
                Debug.println( 2, "EOF" );
                return "EOF";
            }
        switch( next_char ) {
            case (char)-1: {
                Debug.println( 2, "EOF" );
                return "EOF";
                }
            case '|':
                next_char = (char)source.read();
                if( next_char == '|') {
                    next_char = (char)source.read();
                    Debug.print( 2, "||");
                    return "||";
                } else {
                    Debug.print( 2, "|");
                    return "|";
                }
            case '&':
                next_char = (char)source.read();
                Debug.print(2,"&");
                return "&";
            case '@':
                next_char = (char)source.read();
                Debug.print(2,"@");
                return "@";
            case '.':
                next_char = (char)source.read();
                Debug.print(2,".");
                return ".";
            case '*':
                next_char = (char)source.read();
                Debug.print(2,"*");
                return "*";
            case '+':
                next_char = (char)source.read();
                Debug.println(2,"+");
                return "+";
            case ';':
                next_char = (char)source.read();
                Debug.print(2,";");
                return ";";
            case ',':
                next_char = (char)source.read();
                Debug.print(2,",");
                return ",";
            case '\\':
                next_char = (char)source.read();
                Debug.print(2,"\\");
                return "\\";
            case '{':
                next_char = (char)source.read();
                Debug.print(2,"{");
                return "{";
            case '}':
                next_char = (char)source.read();
                Debug.print(2,"}");
                return "}";
            case '(':
                next_char = (char)source.read();
                Debug.print(2,"(");
                return "(";
            case ')':
                next_char = (char)source.read();
                Debug.print(2,")");
                return ")";
	    case '[':
		result = "[";
		next_char = (char)source.read();
		while( next_char != ']' ) {
		    result += next_char;
		    next_char = (char)source.read();
		}
		result += next_char;
		next_char = (char)source.read();
		isTokenID = true;
		Debug.print(2, result);
		return result;
            default:
                if( isLetter(next_char) || next_char == '?' ||
                    next_char == '!' || next_char == '<' || next_char == '#' ) {
                    result = "";
                    if( next_char == '!' || next_char == '?' || next_char == '#') {
                         result = result + next_char;
                         next_char = (char)source.read();
                    }
                    if( next_char == '<' ) {
                        do {
                           result = result + next_char;
                           next_char = (char)source.read();
                        } while( next_char != '>' );
                        result = result + next_char;
                	    next_char = (char)source.read();
		            } else {
                        while( isLetter(next_char) ) {
                        result = result + next_char;
                        next_char = (char)source.read();
                        }
                    }
                    if( next_char != '.' )
		    {
			if( result.equals( "NULL" ) ) 
			{
			    isTokenID = true;
			    Debug.print(2,result);
			    return result;
			}
			throw new SyntaxErrorException("invalide event name: dot delimiter expected in event name, found "+next_char+" after "+result);
                    }
		    else result = result + next_char;
                    next_char = (char)source.read();
                    while( isLetter(next_char) ) {
                       result = result + next_char;
                       next_char = (char)source.read();
                    }
                    if( next_char == '^' || next_char == '$' || next_char == '~' ) {
                            result = result + next_char;
                            next_char = (char)source.read();
                    }
                    isTokenID = true;
                    Debug.print(2,result);
                    return result;
                }
                throw new SyntaxErrorException("Unexpected literal with code "+ (int)next_char);
         }
    }

    public void fillIdentifiers(String prot) {
	ActionToken a;
        try {
    	    boolean incomposition = false;
	    source = new StringReader(prot);
    	    next_char = (char)source.read();
    	    next_token = nextToken();
	    while( next_token != "EOF" ) {
		if( isTokenID && !next_token.equals("NULL") && !next_token.startsWith("[")) 
	    if( next_token.startsWith( "?" ) || next_token.startsWith( "!" ) ||next_token.startsWith( "#" ) ) {
		// this is action /action token
		if( next_token.endsWith( "^" ) ||  next_token.endsWith( "$" ) || next_token.endsWith( "~" ) ) {
		    // this is action token
		    Debug.println(3, "CollectList: action token");
		    a = new ActionToken( next_token );
		    EdgeFactory.getActionTokenIdx(a);
		} else {
		    Debug.println(3, "CollectList: action token without");
		    // this is just action
		    a = new ActionToken( next_token+"^" );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( next_token+"$" ).toInverseActionToken();
		    EdgeFactory.getActionTokenIdx(a);
		}
	    } else {
		// this is event / or event name 
		if( next_token.endsWith( "^" ) ||  next_token.endsWith( "$" ) || next_token.endsWith( "~" ) ) {
		    Debug.println(3, "CollectList: adding event");
		    // this is event
		    a = new ActionToken( "?"+next_token );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( "!"+next_token );
		    EdgeFactory.getActionTokenIdx(a);
		    if( incomposition ) { // this is list for composition
		      a = new ActionToken( "#"+next_token );
 		      EdgeFactory.getActionTokenIdx(a);
		    }
		} else {
		    Debug.println(3, "CollectList: adding event name");
		    // this is event name
		    a = new ActionToken( "?"+next_token+"^" );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( "?"+next_token+"$" );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( "!"+next_token+"^" );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( "!"+next_token+"$" );
		    EdgeFactory.getActionTokenIdx(a);
		    if( incomposition ) { // this is list for composition
		      a = new ActionToken( "#"+next_token+"^" );
		      EdgeFactory.getActionTokenIdx(a);
		      a = new ActionToken( "#"+next_token+"$" );
		      EdgeFactory.getActionTokenIdx(a);
		    }
		}
	    }		
		if( next_token == "&" ) incomposition = !incomposition;
		next_token = nextToken();
	    }
	    Debug.println( 1,"Filling identifiers: finished" );
        } catch( Exception e ) { }
	
    }

    public static void fillOneName(String name, boolean includetau) {
	ActionToken a;
        try {
	    if( name.startsWith( "?" ) || name.startsWith( "!" ) ||name.startsWith( "#" ) ) {
		// this is action /action token
		if( name.endsWith( "^" ) ||  name.endsWith( "$" ) || name.endsWith( "~" ) ) {
		    // this is action token
		    Debug.println(3, "fillOneName: action token");
		    a = new ActionToken( name );
		    EdgeFactory.getActionTokenIdx(a);
		} else {
		    Debug.println(3, "fillOneName: action token without");
		    // this is just action
		    a = new ActionToken( name+"^" );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( name+"$" ).toInverseActionToken();
		    EdgeFactory.getActionTokenIdx(a);
		}
	    } else {
		// this is event / or event name 
		if( name.endsWith( "^" ) ||  name.endsWith( "$" ) || name.endsWith( "~" ) ) {
		    Debug.println(3, "fillOneName: adding event");
		    // this is event
		    a = new ActionToken( "?"+name );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( "!"+name );
		    EdgeFactory.getActionTokenIdx(a);
		    if( includetau ) { // this is list for composition
		      a = new ActionToken( "#"+name );
 		      EdgeFactory.getActionTokenIdx(a);
		    }
		} else {
		    Debug.println(3, "fillOneName: adding event name");
		    // this is event name
		    a = new ActionToken( "?"+name+"^" );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( "?"+name+"$" );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( "!"+name+"^" );
		    EdgeFactory.getActionTokenIdx(a);
		    a = new ActionToken( "!"+name+"$" );
		    EdgeFactory.getActionTokenIdx(a);
		    if( includetau ) { // this is list for composition
		      a = new ActionToken( "#"+name+"^" );
		      EdgeFactory.getActionTokenIdx(a);
		      a = new ActionToken( "#"+name+"$" );
		      EdgeFactory.getActionTokenIdx(a);
		    }
		}
	    }		
        } catch( Exception e ) { }
	
    }

/**
 * Builds protocol for + operator.
 *
 * @author	Stanislav Visnovsky
 * @version	2.0.0
 *
 * @return		built protocol
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see	Builder#sequence
 * @see ProtocolAlternative
*/
    private Protocol alternative() throws SyntaxErrorException, IOException {
        Protocol result = sequence();
        while( next_token == "+" ) {
            next_token = nextToken();
            Protocol m = sequence();
            result = new ProtocolAlternative( result, m );
        }
        return result;
    }

/**
 * Builds protocol for ; operator.
 *
 * @author	Stanislav Visnovsky
 * @version	2.0.0
 *
 * @return		built automaton
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see	Builder#andparallel
 * @see ProtocolSequence
*/
    private Protocol sequence() throws SyntaxErrorException, IOException {
        Protocol result = andparallel();

        while( next_token == ";" ) {
	    next_token = nextToken();
            Protocol m = andparallel();
	    result = new ProtocolSequence( result, m );
        }
	
        return result;
    }

/**
 * Builds protocol for | operator.
 *
 * @author	Stanislav Visnovsky
 * @version	2.0.0
 *
 * @return		built protocol
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see	ProtocolParallel
 * @see Builder#orparallel
 */
    private Protocol andparallel() throws SyntaxErrorException, IOException {
        Protocol result = orparallel();
        while( next_token == "|" ) {
            next_token = nextToken();
            Protocol m = orparallel();
            result = new ProtocolParallel( result, m );
        }
        return result;
    }

/**
 * Builds protocol for || operator.
 *
 * @author	Stanislav Visnovsky
 * @version	2.0.0
 *
 * @return		built protocol
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see	ProtocolOrParallel
 * @see Builder#term
 */
    private Protocol orparallel() throws SyntaxErrorException, IOException {
        Protocol result = term();
        while( next_token == "||" ) {
            next_token = nextToken();
            Protocol m = term();
	    result = new ProtocolOrParallel(result,m);
        }
        return result;
    }

/**
 * Parse a list of comma-separaded ids. Requires that all identifiers to 
 * be already in EdgeFactory.ActionTokens
 *
 * @author	Stanislav Visnovsky
 * @version	2.0.0
 *
 * @return	read list of ids
 * @param endmark   a string to mark the end of the list
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see	Builder#nextToken
 * @see Builder#composition
 * @see Builder#adjustment
 * @see EdgeFactory#ActionTokens
 */
    private ActionTokenArray collectList( String endmark ) throws SyntaxErrorException, IOException {
        ActionTokenArray result = new ActionTokenArray();
        while( isTokenID ) {
	    if( next_token.startsWith( "?" ) || next_token.startsWith( "!" ) ||next_token.startsWith( "#" ) ) {
		// this is action /action token
		if( next_token.endsWith( "^" ) ||  next_token.endsWith( "$" ) || next_token.endsWith( "~" ) ) {
		    // this is action token
		    Debug.println(3, "CollectList: action token");
		    ActionToken a = new ActionToken( next_token );
		    result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		} else {
		    Debug.println(3, "CollectList: action token without");
		    // this is just action
		    ActionToken a = new ActionToken( next_token+"^" );
		    result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		    a = new ActionToken( next_token+"$" ).toInverseActionToken();
		    result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		}
	    } else {
		// this is event / or event name 
		if( next_token.endsWith( "^" ) ||  next_token.endsWith( "$" ) || next_token.endsWith( "~" ) ) {
		    Debug.println(3, "CollectList: adding event");
		    // this is event
		    ActionToken a = new ActionToken( "?"+next_token );
		    result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		    a = new ActionToken( "!"+next_token );
		    result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		    if( endmark == "&" ) { // this is list for composition
		      a = new ActionToken( "#"+next_token );
		      result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		    }
		} else {
		    Debug.println(3, "CollectList: adding event name");
		    // this is event name
		    ActionToken a = new ActionToken( "?"+next_token+"^" );
		    result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		    a = new ActionToken( "?"+next_token+"$" );
		    result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		    a = new ActionToken( "!"+next_token+"^" );
		    result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		    a = new ActionToken( "!"+next_token+"$" );
		    result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		    if( endmark == "&" ) { // this is list for composition
		      a = new ActionToken( "#"+next_token+"^" );
		      result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		      a = new ActionToken( "#"+next_token+"$" );
		      result.tokens[EdgeFactory.getActionTokenIdx(a)]=true;
		    }
		}
	    }		
            next_token = nextToken();
            if( next_token == endmark ) {
                Debug.println( 2,"End mark reached " );
                next_token = nextToken();
                return result;
            }
            if( next_token.compareTo( "," ) != 0) throw new SyntaxErrorException("Comma expected in list");
            else next_token = nextToken();
        }
        if( next_token == endmark ) {
            Debug.println(2, "End mark reached" );
            next_token = nextToken();
            return result;
        }
        else throw new SyntaxErrorException("List syntax error");
    }

/**
 * Builds automaton for restriction operator by invoking the Restrict() operation.
 *
 * @author	Stanislav Visnovsky
 * @version	1.0.0
 *
 * @return		built automaton
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see Machine#Restrict
 * @see Builder#collectList
 */
    private Protocol restriction() throws SyntaxErrorException, IOException {
        Protocol result = alternative();
        while( next_token == "\\" ) {
            Debug.println( 2,"Restriction found");
            next_token = nextToken();
            if( next_token != "(" ) throw new SyntaxErrorException("Restriction list expected");
            next_token = nextToken();
            ActionTokenArray sync = collectList( ")" );
            Debug.println( 2,"Restriction action tokens: "+sync.toList() );
            
	    result.Restrict( new ActionTokenArray(sync ) );
	    
	    result = new ProtocolRestriction( result, sync );
        }
        return result;
    }

/**
 * Builds automaton for composition operator by invoking the createComposition.
 *
 * @author	Stanislav Visnovsky
 * @version	1.0.0
 *
 * @return		built automaton
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see	Builder#alternative
 * @see Builder#createComposition
 * @see Machine
 * @see stateImpl
 */
    private Protocol composition() throws SyntaxErrorException, IOException {
        Protocol result = restriction();
        while( next_token == "&" ) {
            Debug.println(2, "Composition found" );
            next_token = nextToken();
            ActionTokenArray sync = collectList( "&" );
            Debug.println( 2,"Composition action tokens: "+sync.toList() );
            Protocol m = restriction();
            result = new ProtocolComposition( result, m, sync );
        }
        return result;
    }

/**
 * Builds protocol for adjustment operator.
 *
 * @author	Stanislav Visnovsky
 * @version	2.0.0
 *
 * @return		built protocol
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see	Builder#composition
 * @see ProtocolAdjustment
 * @see collectList
 */
    private Protocol adjustment() throws SyntaxErrorException, IOException {
        Protocol result = composition();
        while( next_token == "@" ) {
            Debug.println( 2,"Adjustment found" );
            next_token = nextToken();
            ActionTokenArray sync = collectList( "@" );
            Debug.println( 2,"Adjustment action tokens: "+sync.toList() );
            Protocol m = composition();
            result = new ProtocolAdjustment( result, m, new ActionTokenArray(sync) );
        }
        return result;
    }

/**
 * Builds the automata for unary operator *.
 *
 * @author	Stanislav Visnovsky
 * @version	1.0.0
 *
 * @return		built automaton
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see	Builder#factor
 * @see Machine
 * @see stateImpl
 * @see edgeSimple
*/
    private Protocol term() throws SyntaxErrorException, IOException {
        Protocol result = factor();
        if( next_token == "*" ) {
            next_token = nextToken();
	    result = new ProtocolRepetition( result );
        }
        return result;
    }

/**
 * Builds the basic automata for action tokens and NULL. Also, parentheses are handled here.
 *
 * @author	Stanislav Visnovsky
 * @version	1.0.0
 *
 * @return		built automaton
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
 *
 * @see	Builder#alternative
 * @see Machine
 * @see stateImpl
 * @see edgeSimple
*/
    private Protocol factor() throws SyntaxErrorException, IOException {
        Protocol result;
	if( next_token.equals("NULL") )
	{
	    result = new ProtocolToken( new ActionToken("NULL") );
            next_token = nextToken();
	} else
        if( next_token == "(" ) {
            next_token = nextToken();
            result = adjustment();
            if( next_token != ")" ) { throw new SyntaxErrorException(") expected"); }
            next_token = nextToken();
        } else
        if( next_token.equals("EOF") ) {
            result = null;
        } else { // it has to be an action token or abbreviation
            if( !isTokenID ) throw new SyntaxErrorException("identifier expected, last token "+next_token);
	    String event = next_token;
            next_token = nextToken();
	    if( next_token == "{" ) {
        	// nested call abbreviation
        	next_token = nextToken();
        	Protocol pom = alternative();

        	if( next_token != "}" ) throw new SyntaxErrorException("} expected");
                    next_token = nextToken();

        	if( !event.startsWith( "?" ) ) event = "?" + event;
        	ActionToken start = new ActionToken(event+"^");

            	ActionToken stop = start.toInverseActionToken();

            	stop = new ActionToken( stop.name.substring(0,stop.name.length()-1) + "$" );

		pom = new ProtocolSequence(  new ProtocolToken(  start ), pom );
		pom = new ProtocolSequence( pom, new ProtocolToken( stop ) );
		
		Debug.println(3, "Nested call abbreviation:");
		((Printer)pom).Print(3);
		
        	result = pom;
	    } else {
	// action token or simple call abbrevation
    		ActionToken a = new ActionToken( event );
		if( a.isSimpleCall() ) {	// simple call abbreviation
            	    ActionToken start = new ActionToken( a.name+'^') ;
            	    ActionToken stop = new ActionToken( a.name+'$').toInverseActionToken();
		    result = new ProtocolSequence( new ProtocolToken( start ), new ProtocolToken( stop ) );
		} else {	// action token
		    result = new ProtocolToken( a );
		}
	    }
        }
        return result;
    }

/**
 * The main entry point for building the automaton for a behavior protocol. Using all
 * private methods of this class, this method takes a string as the argument and then it
 * builds the automton by invoking Build for reader argument
 *
 * @author	Stanislav Visnovsky
 * @version	1.0.0
 *
 * @param protocol	a reader containing a protocol to be build
 * @return		built automaton
 * @throws SyntaxErrorException in case of badly formed protocols
 * @throws IOException		in case of unexpected end of reader
*/
    public Protocol Build( String protocol ) throws SyntaxErrorException {
        Debug.println( 1,"Protocol Build: start");
        try {
    	    source = new StringReader(protocol);
    	    next_char = (char)source.read();
    	    next_token = nextToken();
    	    Protocol m = adjustment();
	    Debug.println( 1,"Protocol Build: finished" );
	    ((Printer)m).Print( 2 );
    	    return m;
        } catch( IOException e ) { return null; }
    }

}